---------------------------------------------------------------
-- Temporal data in SQL Server 
---------------------------------------------------------------

---------------------------------------------------------------
-- PART 1: optimization problem, and enhanced T-SQL solution
---------------------------------------------------------------

---------------------------------------------------------------
-- Demo table with application time
---------------------------------------------------------------

-- Creating a demo table
USE tempdb;
GO
SELECT OrderLineID AS id,
 StockItemID * (OrderLineID % 5 + 1) AS b,
 LastEditedBy + StockItemID * (OrderLineID % 5 + 1) AS e
INTO dbo.Intervals
FROM WideWorldImporters.Sales.OrderLines;
-- 231412 rows
GO
ALTER TABLE dbo.Intervals ADD CONSTRAINT PK_Intervals PRIMARY KEY(id);
CREATE INDEX idx_b ON dbo.Intervals(b) INCLUDE(e);
CREATE INDEX idx_e ON dbo.Intervals(e) INCLUDE(b);
GO

-- Checking the data
SELECT *
FROM dbo.Intervals;
SELECT MIN(b), MAX(e)
FROM dbo.Intervals;
-- 1, 1155
GO

-- Date numbers table
CREATE TABLE dbo.DateNums
 (n INT NOT NULL PRIMARY KEY,
  d DATE NOT NULL);
GO
DECLARE @i AS INT = 1, 
 @d AS DATE = '20140701';
WHILE @i <= 1200
BEGIN
INSERT INTO dbo.DateNums
 (n, d)
SELECT @i, @d;
SET @i += 1;
SET @d = DATEADD(day,1,@d);
END;
GO

-- Giving intervals the context
SELECT i.id,
 i.b, d1.d AS dateB,
 i.e, d2.d AS dateE
FROM dbo.Intervals AS i
 INNER JOIN dbo.DateNums AS d1
  ON i.b = d1.n
 INNER JOIN dbo.DateNums AS d2
  ON i.e = d2.n
ORDER BY i.id;
GO

---------------------------------------------------------------
-- Optimizing a query for overlapping intervals - the problem
---------------------------------------------------------------

-- Intervals at the beginning or the end of the timeline 
-- Fast query
SET STATISTICS IO ON;
DECLARE @b AS INT = 10,
 @e AS INT = 30;
SELECT id, b, e
FROM dbo.Intervals
WHERE b <= @e
  AND e >= @b
OPTION (RECOMPILE);
GO
-- 36 logical reads

-- Intervals in the middle of the timeline 
-- Slow query
DECLARE @b AS INT = 570,
 @e AS INT = 590;
SELECT id, b, e
FROM dbo.Intervals
WHERE b <= @e
  AND e >= @b
OPTION (RECOMPILE);
GO
-- 111 logical reads

---------------------------------------------------------------
-- Enhanced T-SQL solution
---------------------------------------------------------------

-- Intervals in the middle of the timeline 
-- Enhanced query
DECLARE @b AS INT = 570,
 @e AS INT = 590;
DECLARE @max AS INT = 20;
SELECT id, b, e
FROM dbo.Intervals
WHERE b <= @e AND b >= @b - @max
  AND e >= @b AND e <= @e + @max
OPTION (RECOMPILE);
-- 20 logical reads
GO

---------------------------------------------------------------
-- Clean up
---------------------------------------------------------------

DROP TABLE dbo.DateNums;
DROP TABLE dbo.Intervals;
GO


---------------------------------------------------------------
-- PART 2: SQL Server system-versioned tables
---------------------------------------------------------------

---------------------------------------------------------------
-- WideWorldImporters temporal tables
---------------------------------------------------------------

USE WideWorldImporters;
GO

-- For SYSTEM_TIME ALL gets all current and historical rows
SELECT COUNT(*) AS CurrentCountries
FROM Application.Countries;
-- 190

SELECT COUNT(*) AS ArchiveCountries
FROM Application.Countries_Archive;
-- 36

SELECT COUNT(*) AS CurrrentAndArchiveCountries
FROM Application.Countries
 FOR SYSTEM_TIME ALL;
-- 226

-- Different dates
SELECT DISTINCT ValidFrom, ValidTo
FROM Application.Countries
 FOR SYSTEM_TIME ALL
ORDER BY ValidFrom, ValidTo;

-- All rows
SELECT *
FROM Application.Countries
 FOR SYSTEM_TIME ALL;
-- 226

-- State on a specific time point
SELECT *
FROM Application.Countries
 FOR SYSTEM_TIME AS OF '2013-01-02 00:00:00.0000000';
-- 190

-- Current and historical rows, no rows on upper boundary
SELECT *
FROM Application.Countries
 FOR SYSTEM_TIME FROM '2013-01-01 00:00:00.0000000' TO '2013-07-01 16:00:00.0000000';
-- 190 (no historical rows included)

-- Current and historical rows, also rows on upper boundary
SELECT *
FROM Application.Countries
 FOR SYSTEM_TIME BETWEEN '2013-01-01 00:00:00.0000000' AND '2013-07-01 16:00:00.0000000';
-- 201

-- Rows with activity in the interval, including boundaries
SELECT *
FROM Application.Countries
 FOR SYSTEM_TIME CONTAINED IN ('2013-01-01 00:00:00.0000000', '2013-07-01 16:00:00.0000000');
GO

---------------------------------------------------------------
-- SQL Server temporal tables as DW dimensions
---------------------------------------------------------------

-- 2. DimEmployee
SELECT *
FROM Application.People
WHERE IsEmployee = 1;
-- 19 current rows

SELECT *
FROM Application.People
 FOR SYSTEM_TIME ALL
WHERE IsEmployee = 1;
-- 211 rows with history

-- Query to populate the DimEmployee dimension with all of the history
-- Check the data
SELECT PersonID AS WWIEmployeeID,
 FullName AS EmployeeName,
 IsSalesperson AS SalesPerson,
 ValidFrom,
 ValidTo,
 IIF (ValidTo = '9999-12-31 23:59:59.9999999', 1, 0) AS CurrentRecord,
 1 AS LineageKey
FROM Application.People
 FOR SYSTEM_TIME ALL
WHERE IsEmployee = 1
  AND PersonID = 14;
GO
-- Lily Code has multiple records for a single day
-- For example, 2013-01-01 00:00:00.0000000 - 2013-01-01 08:00:00.0000000
-- and          2013-01-01 08:00:00.0000000 - 2013-01-19 08:00:00.0000000
-- Must get the last row per day and make closed time intervals
-- Also lower the ValidTo for a day, except for 9999-12-31

WITH PersonCTE AS 
(
SELECT PersonID AS WWIEmployeeID,
 FullName AS EmployeeName,
 IsSalesperson AS SalesPerson,
 CAST(ValidFrom AS DATE) AS ValidFrom,
 IIF (ValidTo = '9999-12-31 23:59:59.9999999',
      CAST(ValidTo AS DATE),
      DATEADD(day, -1, CAST(ValidTo AS DATE))) AS ValidTo,
 ROW_NUMBER() OVER(PARTITION BY PersonID, CAST(ValidFrom AS Date)
                 ORDER BY ValidFrom DESC) AS rn,
 IIF (ValidTo = '9999-12-31 23:59:59.9999999', 1, 0) AS CurrentRecord,
 1 AS LineageKey
FROM Application.People
 FOR SYSTEM_TIME ALL
WHERE IsEmployee = 1
)
SELECT WWIEmployeeID,
 EmployeeName,
 SalesPerson,
 ValidFrom,
 ValidTo,
 CurrentRecord,
 LineageKey
FROM PersonCTE
WHERE rn = 1;
GO
-- 190 rows


---------------------------------------------------------------
-- Temporal data in SQL Server end
---------------------------------------------------------------
